#define IN_MODULE

#include "../../include/aftersteplib.h"
#include "../../include/afterstep.h"
#include "../../include/menus.h"
#include "../../include/style.h"
#include "../../include/screen.h"
#include "../../include/parser.h"
#include "../../include/confdefs.h"
#include "../../include/mystyle.h"

#undef StickyIcons

/*#define DEBUG_PAGER */
/*#define DEBUG_X_PAGER */
/*#define DEBUG_BACKGROUNDS*/


typedef struct pager_window 
{
  Window w;
  Window frame;
  int x;
  int y;
  int width;
  int height;
  int desk;
  int frame_x;
  int frame_y;
  int frame_width;
  int frame_height;
  int title_height;
  int border_width;
  int icon_x;
  int icon_y;
  int icon_width;
  int icon_height;
  unsigned long flags;
  Window icon_w;
  Window icon_pixmap_w;
  char *icon_name;
  Window PagerView;        /* window to be shown in Pager desk window */
  Window IconView;         /* window to be shown in iconized Pager */

  struct pager_window *next;
} PagerWindow;

typedef struct pager_view_position
{
  int normal_x, normal_y, normal_width, normal_height ;
  int icon_x, icon_y, icon_width, icon_height ;
}PagerViewPosition ;

typedef enum pager_backgr_type
{
  backgrUseXPM = 0,
  backgrScaleRoot,
  backgrSameAsRoot,
  backgrTransparent

} BackgroundType;

typedef struct filed_image
{
  Pixmap pixmap ;
  char*  filename;
  char*  LoaderArgs ;  
  int    in_use ;        /*number of objects using it*/
} FiledImage ;


typedef struct desk_info 
{
  Window w;
  Window title_w;

  int x;
  int y;
  char *label;
  
  BackgroundType PagerBackType ;
  int RootBack, PagerBack ;
  char *StyleName ;
  Pixel TintColor ;
  GC TintGC ;
  
  char* LoaderArgs ;

  int FastBackgroundOn ;
  unsigned int tile_hsize;
  unsigned int tile_vsize;
  
} DeskInfo;

typedef struct pager_info
{
  Window Pager_w;

  unsigned short Flags ;

#ifdef PAGER_BACKGROUND

#define BACKS_NUM                        ((Pager.ndesks*2)+1)
#define SMALL_FILE_NAME_EXT              ".%ux%u.as.xpm"
#define SMALL_FILE_NAME_EXT_MAX_LEN      32

  char* XImageLoaderApp ;
  char* XImageLoaderArgs ;

  FiledImage* Backgrounds ;
#endif

  DeskInfo *Desks;
  Window SelectionWin[4];
  PagerWindow *Start;
  PagerWindow *FocusWin;

  int desk1, desk2, ndesks;
  int Rows, Columns;
  int PageRows, PageColumns ;
  int xSize, ySize ; /* x and y size of desktop */
  
  int bStarted ; 
  
  Pixmap CurentRootBack ;

} PagerInfo;


typedef enum {
    WIN_FOCUSED = 0,
    WIN_STICKY,
    WIN_UNFOCUSED,
    WIN_TYPES_NUM
}PagerWinType;

typedef struct pager_win_attr {
    Pixel  back_pixel;
    Pixmap back_pixmap;
    GC     foreGC ;
}PagerWinAttr;    

typedef enum {
    STYLE_FWINDOW=WIN_FOCUSED,	
    STYLE_SWINDOW=WIN_STICKY,	
    STYLE_UWINDOW=WIN_UNFOCUSED, 
    STYLE_ADESK,	
    STYLE_INADESK,
    STYLE_MAX_STYLE
} PAGER_STYLES ;

typedef struct pager_look {

  MyStyle* Styles[STYLE_MAX_STYLE];
  GC WinForeGC[WIN_TYPES_NUM];
  MyStyle** DeskStyles ;
   
  int DeskBorderWidth ;
  int TitleAlign ;

  Pixel GridColor ;
  GC	GridGC ;

  Pixel BorderColor ;  
  Pixel SelectionColor ; 

  MyFont windowFont ; 
  int bNeedToFixStyles ;

}PagerLook ;

typedef struct pager_pipes {
    int fd[2];
    int fd_width ;
    int x_fd ;
} PagerPipes ;


typedef struct pager_atom {
    Atom atom;
    char* name;
    Atom type;
}PagerAtom ;

/* global constants definitions */

extern PagerInfo Pager ;
extern PagerLook Look;
extern PagerPipes Pipes;

typedef enum {
    ATOM_WM_DEL_WIN=0,
    ATOM_ROOT_PIXMAP,
    ATOM_STYLES,
    ATOMS_NUM
} PagerAtoms ;
extern PagerAtom  Atoms[];

/*extern PagerWinAttr WinAttributes[WIN_TYPES_NUM] ;*/

#ifdef PAGER_BACKGROUND
#define PAGER_PIXMAP     Pager.Backgrounds[Pager.Desks[Desk].PagerBack].pixmap
#define ROOT_PIXMAP      Pager.Backgrounds[Pager.Desks[Desk].RootBack].pixmap

#define PAGER_FILENAME   Pager.Backgrounds[Pager.Desks[Desk].PagerBack].filename
#define ROOT_FILENAME    Pager.Backgrounds[Pager.Desks[Desk].RootBack].filename
#endif

#define ON  1
#define OFF 0

/*************************************************************************
 *
 * Subroutine Prototypes
 * 
 *************************************************************************/
void SendInfo(int *fd,char *message,unsigned long window);
void DeadPipe(int nonsense);
void process_message(unsigned long type,unsigned long *body);
void GetOptions(const char *filename);
void GetBaseOptions(const char *filename);

PagerWindow *FindWindow (Window w);
void DeleteWindows( PagerWindow*t );
void list_add(unsigned long *body);
void list_configure(unsigned long *body);
void list_destroy(unsigned long *body);
void list_focus(unsigned long *body);
void list_toggle(unsigned long *body);
void list_new_page(unsigned long *body);
void list_new_desk(unsigned long *body);
void list_new_background(unsigned long *body);
void list_raise(unsigned long *body);
void list_lower(unsigned long *body);
void list_unknown(unsigned long *body);
void list_iconify(unsigned long *body);
void list_deiconify(unsigned long *body);
void list_window_name(unsigned long *body);
void list_icon_name(unsigned long *body);
void list_class(unsigned long *body);
void list_res_name(unsigned long *body);
void list_end(void);
int My_XNextEvent( XEvent *event);

/* look management stuff */
void LookCleanUp();
void GetLook( FILE* fd );
void FixLook();
void ApplyLook();
void OnLookUpdated();

/* Stuff in background.c */
int DoPagerTinting(long Desk);
void SetPagerBackground(long Desk);
#ifdef PAGER_BACKGROUND
void InitXImageLoader();
void SetXImageLoaderArgs(char* args);
void InitDeskBackgrounds( long Desk );
void InitBackgrounds();
void MakeBackgrounds( long Desk );
void CheckAndGrabRootPixmap( long Desk );
int IfRootPixmap( long Desk );
int SetERootPixmap( long Desk );
void DrawRoot( long Desk );
void BackgroundChanged( long Desk );
void SetBackgroundType(long Desk, char* params );
void SetRootPixmap(long Desk, char *xpmFile);
void SetPagerPixmap(long Desk, char *xpmFile);
int ScalePagerBackground( long Desk );

#endif

/* Stuff in x_pager.c */
void InitDesk( long Desk );
void initialize_pager(void);
Pixel GetColor(char *name);
void nocolor(char *a, char *b);
void DispatchEvent(XEvent *Event);
void DecorateDesk (long Desk);
void UpdateTransparency();
void DecoratePager();
void DisplayFrame( long Desk );
void LowerFrame();
void HideFrame();
void DestroyView( PagerWindow* t );
void ReConfigure(void);
void ReConfigureEx( unsigned width, unsigned height );
void ReConfigureAll(void);
void MovePage(void);
void DrawGrid(int i);
void LabelDesk(int i  /*Desk */ );
void HilightDesk(int i, int if_texture);
void DrawIconGrid(int erase);
void SwitchToDesk(int Desk);
void SwitchToDeskAndPage(int Desk, XEvent *Event);
int  GetWinAttributes( PagerWindow* t, 
                       XSetWindowAttributes* attributes, 
		       unsigned long *valuemask, 
		       PagerViewPosition* ppos );
void AddNewWindow(PagerWindow *prev);
void MoveResizePagerView(PagerWindow *t);
void ChangeDeskForWindow(PagerWindow *t,long newdesk);
void MoveStickyWindow(void);
void MapBalloonWindow(XEvent *);
void UnmapBalloonWindow(void);
void DrawInBalloonWindow(void);
void Hilight(PagerWindow *t);
void Scroll(int Desk, int x, int y);
void MoveWindow(XEvent *Event);

PagerWinType GetWinType(PagerWindow * t);
void PutLabel(PagerWindow * t, Window w);
#define LabelWindow(t) 		PutLabel(t,t->PagerView )
#define LabelIconWindow(t) 	PutLabel(t,t->IconView )

void ReConfigureIcons(void);
void IconSwitchPage(XEvent *Event);
void IconScroll(int x, int y);
void IconMoveWindow(XEvent *Event,PagerWindow *t);
void HandleExpose(XEvent *Event);
void MoveStickyWindows(void);


#ifdef BROKEN_SUN_HEADERS
#include "../../afterstep/sun_headers.h"
#endif


